import os

import grpc
from rich.console import Console

err_console = Console(stderr=True)


class ErrorPrinter:
    @staticmethod
    def print_grpc_error(error: grpc.RpcError):
        if error.code() == grpc.StatusCode.UNAUTHENTICATED:
            is_access_token = os.environ["ACCESS_TOKEN"]
            if is_access_token is not None and is_access_token == "true":
                err_console.print(f":boom: [bold red]Authentication failed[/bold red]: {error.details()}")
                err_console.print("Please login again")
            else:
                err_console.print(":boom: [bold red]Authentication failed[/bold red]")
                err_console.print("Failed to verify api-key")
        else:
            # print(f"Unexpected error: {error.code()}")
            err_console.print(f":boom: [bold red]Unexpected error, status code[/bold red]: {error.code()}")
            err_console.print(error.details())
        exit(1)

    @staticmethod
    def print_hint(message: str):
        err_console.print(f":point_right: [bold]{message}[/bold]")

    @staticmethod
    def print_generic_error(message: str):
        err_console.print(f":boom: [bold red]Failed[/bold red]: {message}")

    @staticmethod
    def print_generic_message(message: str):
        err_console.print(f"[bold]{message}[/bold]:")



from __future__ import annotations

import json
import tempfile

import grpc
import typer
from rich.progress import Progress, SpinnerColumn, TextColumn

from cli.errors import ErrorPrinter

from ..broker.lib.broker import Broker
from . import rest_helper as rest

app = typer.Typer(
    help="""
Support for playback of a recording on a cloud broker, make sure to always mount a recording first
"""
)


@app.command()
def play(
    recording_session: str = typer.Argument(..., help="Recording session id", envvar="REMOTIVE_CLOUD_RECORDING_SESSION"),
    broker: str = typer.Option(None, help="Broker to use"),
    project: str = typer.Option(..., help="Project ID", envvar="REMOTIVE_CLOUD_PROJECT"),
    show_progress: bool = typer.Option(False, help="Show progress after started playing"),
):
    """
    Start playing a recording
    """
    _do_change_playback_mode("play", recording_session, broker, project)
    if progress:
        _do_change_playback_mode("status", recording_session, broker, project)


@app.command()
def pause(
    recording_session: str = typer.Argument(..., help="Recording session id", envvar="REMOTIVE_CLOUD_RECORDING_SESSION"),
    broker: str = typer.Option(None, help="Broker to use"),
    project: str = typer.Option(..., help="Project ID", envvar="REMOTIVE_CLOUD_PROJECT"),
):
    """
    Pause a recording
    """
    _do_change_playback_mode("pause", recording_session, broker, project)


@app.command()
def progress(
    recording_session: str = typer.Argument(..., help="Recording session id", envvar="REMOTIVE_CLOUD_RECORDING_SESSION"),
    broker: str = typer.Option(None, help="Broker to use"),
    project: str = typer.Option(..., help="Project ID", envvar="REMOTIVE_CLOUD_PROJECT"),
):
    """
    Shows progress of the recording playing
    """
    _do_change_playback_mode("status", recording_session, broker, project)


@app.command()
def seek(
    recording_session: str = typer.Argument(..., help="Recording session id", envvar="REMOTIVE_CLOUD_RECORDING_SESSION"),
    seconds: int = typer.Option(..., min=0, help="Target offset in seconds"),
    broker: str = typer.Option(None, help="Broker to use"),
    project: str = typer.Option(..., help="Project ID", envvar="REMOTIVE_CLOUD_PROJECT"),
):
    """
    Seek seconds into a recording
    """
    _do_change_playback_mode("seek", recording_session, broker, project, seconds)


@app.command()
def stop(
    recording_session: str = typer.Argument(..., help="Recording session id", envvar="REMOTIVE_CLOUD_RECORDING_SESSION"),
    broker: str = typer.Option(None, help="Broker to use"),
    project: str = typer.Option(..., help="Project ID", envvar="REMOTIVE_CLOUD_PROJECT"),
):
    """
    Stop playing
    """
    _do_change_playback_mode("stop", recording_session, broker, project)


def _do_change_playback_mode(mode: str, recording_session: str, broker: str, project: str, seconds: int | None = None):  # noqa: C901
    response = rest.handle_get(f"/api/project/{project}/files/recording/{recording_session}", return_response=True)
    r = json.loads(response.text)
    recordings: list = r["recordings"]
    files = list(map(lambda rec: {"recording": rec["fileName"], "namespace": rec["metadata"]["namespace"]}, recordings))

    if broker is not None:
        response = rest.handle_get(f"/api/project/{project}/brokers/{broker}", return_response=True, allow_status_codes=[404])
    else:
        response = rest.handle_get(f"/api/project/{project}/brokers/personal", return_response=True, allow_status_codes=[404])
    if response.status_code == 404:
        broker_arg = ""
        if broker is not None:
            broker_arg = f" --broker {broker} --ensure-broker-started"
        ErrorPrinter.print_generic_error("You need to mount the recording before you play")
        ErrorPrinter.print_hint(f"remotive cloud recordings mount {recording_session}{broker_arg} --project {project}")
        exit(1)

    broker_info = json.loads(response.text)
    broker = Broker(broker_info["url"], None)
    try:
        # Here we try to verify that we are operating on a recording that is mounted on the
        # broker so we can verify this before we try playback and can also present some good
        # error messages
        tmp = tempfile.NamedTemporaryFile()
        broker.download(".cloud.context", tmp.name, True)
        with open(tmp.name, "r") as f:
            json_context = json.loads(f.read())
            if json_context["recordingSessionId"] != recording_session:
                ErrorPrinter.print_generic_error(
                    f"The recording id mounted is '{json_context['recordingSessionId']}' "
                    f"which  not the same as you are trying to {mode}, "
                    "use cmd below to mount this recording"
                )
                ErrorPrinter.print_hint(f"remotive cloud recordings mount {recording_session} --project {project}")
                exit(1)
    except grpc.RpcError as rpc_error:
        if rpc_error.code() == grpc.StatusCode.NOT_FOUND:
            ErrorPrinter.print_generic_error(f"You must use mount to prepare a recording before you can use {mode}")
            ErrorPrinter.print_hint(f"remotive cloud recordings mount {recording_session} --project {project}")
        else:
            ErrorPrinter.print_grpc_error(rpc_error)
        exit(1)
    if mode == "pause":
        broker.pause_play(files, True)
    elif mode == "play":
        r = broker.play(files, True)
    elif mode == "seek":
        broker.seek(files, int(seconds * 1000000), True)
    elif mode == "stop":
        broker.seek(files, 0, True)
    elif mode == "status":
        p = Progress(SpinnerColumn(), TextColumn("[progress.description]{task.description}"), transient=True)
        t = p.add_task("label", total=1)

        with p:
            broker.listen_on_playback(lambda c: p.update(t, description=str(c)))
    else:
        raise Exception(f"Illegal command {mode}")



from __future__ import annotations

import sys
from pathlib import Path

import typer
from typing_extensions import List

from cli.broker.lib.broker import SubscribableSignal
from cli.errors import ErrorPrinter

app = typer.Typer()


@app.command()
def protopie(
    config: Path = typer.Option(
        None,
        exists=True,
        file_okay=True,
        dir_okay=False,
        writable=False,
        readable=True,
        resolve_path=True,
        help="Configuration file with signal subscriptions and mapping if needed",
    ),
    signal: List[str] = typer.Option(None, help="Signal names to subscribe to, mandatory when not using script"),
    signal_name_expression: str = typer.Option(
        None, help='[Experimental] Python expression to rename signal names, i.e \'lower().replace(".","_")\''
    ),
    broker_url: str = typer.Option(..., help="Broker URL", envvar="REMOTIVE_BROKER_URL"),
    api_key: str = typer.Option(None, help="Cloud Broker API-KEY", envvar="REMOTIVE_BROKER_API_KEY"),
    pp_connect_host: str = typer.Option("http://localhost:9981", help="ProtoPie Connect URL"),
):
    """
    ProtoPie Connect bridge-app to connect signals with RemotiveBroker

    Subscribe to signals and send signal values to your Pie in a simple way. You can subscribe to signals from command line
    using --signal or use --config to use a json configuration file.

    ```
    $ remotive connect protopie --signal vss:Vehicle.Chassis.SteeringWheel.Angle --signal vss:Vehicle.Speed
    ```

    You can use a configuration file for this if you have many signals, want to share the configuration or you want
    custom mapping of the signal name.
    ```
    $ remotive connect protopie --config my-protopie-config.json
    ```

    Sample my-protopie-config.json
    ```
    {
      "subscription": {
        "Vehicle.CurrentLocation.Heading": {
          "namespace": "vss"
        },
        "Vehicle.Speed": {
          "namespace": "vss",
           "mapTo": ["Speed", "VehicleSpeed"]
        }
      }
    }
    ```
    For simple changes to signal names its possible to use a simple python expression that will be applied to all signal
    names before its published to ProtoPie connect, i.e replacing chars or substring to match variables in Pie. This is
    intended for simple use cases when you do not have a configuration file, otherwise we recommend using the mapTo
    field in the configuration.

    This will replace all occurrences of . (dot) with _ (underscore)
        ```
    $ remotive connect protopie --signal vss:Vehicle.Chassis.SteeringWheel.Angle --signal-name-expression 'replace(".", "_")'
    ```
    """
    from .protopie import protopie

    if len(signal) > 0 and config is not None:
        ErrorPrinter.print_hint("You must choose either --signal or --config, not both")
        sys.exit(1)

    if len(signal) == 0 and config is None:
        ErrorPrinter.print_hint("You must choose either --signal or --config")
        sys.exit(1)

    def to_subscribable_signal(sig: str):
        arr = sig.split(":")

        if len(arr) != 2:
            ErrorPrinter.print_hint(f"--signal must have format namespace:signal ({sig})")
            sys.exit(1)

        return SubscribableSignal(namespace=arr[0], name=arr[1])

    if len(signal) > 0:
        signals_to_subscribe_to = list(map(to_subscribable_signal, signal))
    else:
        signals_to_subscribe_to = None

    protopie.do_connect(
        address=pp_connect_host,
        broker_url=broker_url,
        api_key=api_key,
        expression=signal_name_expression,
        config=config,
        signals=signals_to_subscribe_to,
    )



from __future__ import annotations

from typing import List

import grpc
import typer

from cli.errors import ErrorPrinter

from .lib.broker import Broker

app = typer.Typer(help=help)


@app.command()
def start(
    filename: str = typer.Argument(..., help="Path to local file to upload"),
    namespace: List[str] = typer.Option(..., help="Namespace to record"),
    url: str = typer.Option(..., help="Broker URL", envvar="REMOTIVE_BROKER_URL"),
    api_key: str = typer.Option("offline", help="Cloud Broker API-KEY or access token", envvar="REMOTIVE_BROKER_API_KEY"),
):
    try:
        broker = Broker(url, api_key)
        broker.record_multiple(namespace, filename)
    except grpc.RpcError as rpc_error:
        ErrorPrinter.print_grpc_error(rpc_error)


@app.command()
def stop(
    filename: str = typer.Argument(..., help="Path to local file to upload"),
    namespace: List[str] = typer.Option(..., help="Namespace to record"),
    url: str = typer.Option(..., help="Broker URL", envvar="REMOTIVE_BROKER_URL"),
    api_key: str = typer.Option("offline", help="Cloud Broker API-KEY or access token", envvar="REMOTIVE_BROKER_API_KEY"),
):
    try:
        broker = Broker(url, api_key)
        broker.stop_multiple(namespace, filename)
        print("Successfully stopped recording")
    except grpc.RpcError as rpc_error:
        ErrorPrinter.print_grpc_error(rpc_error)



from __future__ import annotations

import json
import os
import time
from pathlib import Path
from typing import Dict, List, Union

import grpc
import socketio
from remotivelabs.broker.sync import BrokerException, Client, SignalsInFrame
from rich import print as pretty_print
from rich.console import Console
from socketio.exceptions import ConnectionError as SocketIoConnectionError

from cli import settings
from cli.broker.lib.broker import SubscribableSignal
from cli.errors import ErrorPrinter

global PP_CONNECT_APP_NAME
PP_CONNECT_APP_NAME = "RemotiveBridge"

io = socketio.Client()

err_console = Console(stderr=True)

_has_received_signal = False


@io.on("connect")
def on_connect():
    print("Connected to ProtoPie Connect")
    io.emit("ppBridgeApp", {"name": PP_CONNECT_APP_NAME})
    io.emit("PLUGIN_STARTED", {"name": PP_CONNECT_APP_NAME})

    global is_connected
    is_connected = True


# TODO - Receive message from ProtoPie connect


def get_signals_and_namespaces(
    config: Union[Path, None] = None, signals_to_subscribe_to: Union[List[SubscribableSignal], None] = None
) -> (List[str], List[str], Union[Dict[str, str], None]):
    if config is not None:
        with open(config) as f:
            mapping = json.load(f)
            sub = mapping["subscription"]
            signals = list(sub.keys())
            namespaces = list(map(lambda x: sub[x]["namespace"], signals))
    else:
        signals = list(map(lambda s: s.name, signals_to_subscribe_to))
        namespaces = list(map(lambda s: s.namespace, signals_to_subscribe_to))
        sub = None
    return signals, namespaces, sub


def get_signal_name(expression: str, s_name: str) -> str:
    if expression is not None:
        try:
            sig_name = eval(f"s_name.{expression}")
            return sig_name
        except Exception as e:
            ErrorPrinter.print_generic_error(f"Failed to evaluate your python expression {expression}")
            err_console.print(e)
            # This was the only way I could make this work, exiting on another thread than main
            os._exit(1)
    else:
        return s_name


def _connect_to_broker(
    config: Union[Path, None] = None,
    signals_to_subscribe_to: Union[List[SubscribableSignal], None] = None,
    expression: str = None,
):  # noqa: C901
    signals, namespaces, sub = get_signals_and_namespaces(config, signals_to_subscribe_to)

    def on_signals(frame: SignalsInFrame):
        global _has_received_signal
        if not _has_received_signal:
            pretty_print("Bridge-app is properly receiving signals, you are good to go :thumbsup:")
            _has_received_signal = True

        for s in frame:
            if config is not None:
                sig = sub[s.name()]
                sig = s.name() if "mapTo" not in sig.keys() else sig["mapTo"]
                if isinstance(sig, list):
                    for ss in sig:
                        io.emit("ppMessage", {"messageId": get_signal_name(expression, ss), "value": str(s.value())})
                else:
                    io.emit("ppMessage", {"messageId": get_signal_name(expression, sig), "value": str(s.value())})
            else:
                signal_name = get_signal_name(expression, s.name())
                io.emit("ppMessage", {"messageId": signal_name, "value": str(s.value())})

    grpc_connect(on_signals, signals, namespaces)


def grpc_connect(on_signals, signals, namespaces):
    try:
        pretty_print("Connecting and subscribing to broker...")
        subscription = None
        client = Client(client_id="cli")
        client.connect(url=broker, api_key=x_api_key)
        client.on_signals = on_signals

        subscription = client.subscribe(signal_names=signals, namespaces=namespaces, changed_values_only=False)
        pretty_print("Subscription to broker completed")
        pretty_print("Waiting for signals...")

        while True:
            time.sleep(1)

    except grpc.RpcError as e:
        err_console.print(":boom: [red]Problems connecting or subscribing[/red]")
        if isinstance(e, grpc.Call):
            print(f"{e.code()} - {e.details()}")
        else:
            print(e)

    except BrokerException as e:
        print(e)
        if subscription is not None:
            subscription.cancel()

    except KeyboardInterrupt:
        print("Keyboard interrupt received. Closing subscription.")
        if subscription is not None:
            subscription.cancel()

    except Exception as e:
        err_console.print(f":boom: {e}")
        # exit(1)


def do_connect(
    address: str,
    broker_url: str,
    api_key: Union[str, None],
    config: Union[Path, None],
    signals: List[SubscribableSignal],
    expression: Union[str, None],
):
    global broker
    global x_api_key
    global config_path
    broker = broker_url

    if broker_url.startswith("https"):
        if api_key is None:
            print("No --api-key, reading token from file")
            x_api_key = settings.read_token()
        else:
            x_api_key = api_key
    else:
        x_api_key = api_key
    try:
        io.connect(address)
        config_path = config
        while is_connected is None:
            time.sleep(1)
        _connect_to_broker(signals_to_subscribe_to=signals, config=config, expression=expression)
    except SocketIoConnectionError as e:
        err_console.print(":boom: [bold red]Failed to connect to ProtoPie Connect[/bold red]")
        err_console.print(e)
        exit(1)
    except Exception as e:
        err_console.print(":boom: [bold red]Unexpected error[/bold red]")
        err_console.print(e)
        exit(1)



import json
import os
import sys
from pathlib import Path

import typer

from . import rest_helper as rest

app = typer.Typer()

token_file_name = str(Path.home()) + "/.config/.remotive/cloud.secret.token"
config_dir_name = str(Path.home()) + "/.config/.remotive/"


@app.command(name="create", help="Create and download a new personal access token")
def get_personal_access_token(activate: bool = typer.Option(False, help="Activate the token for use after download")):
    rest.ensure_auth_token()
    response = rest.handle_post(url="/api/me/keys", return_response=True)

    if response.status_code == 200:
        name = response.json()["name"]
        path_to_file = write_personal_token(f"personal-token-{name}.json", response.text)
        print(f"Personal access token written to {path_to_file}")
        if not activate:
            print(f"Use 'remotive cloud auth tokens activate {os.path.basename(path_to_file)}' to use this access token from cli")
        else:
            do_activate(path_to_file)
            print("Token file activated and ready for use")
        print("\033[93m This file contains secrets and must be kept safe")
    else:
        print(f"Got status code: {response.status_code}")
        print(response.text)


@app.command(name="list", help="List personal access tokens")
def list_personal_access_tokens():
    rest.ensure_auth_token()
    rest.handle_get("/api/me/keys")


@app.command(name="revoke", help="Revoke the specified access token")
def revoke(name: str = typer.Option(..., help="Name of the access token to revoke")):
    rest.ensure_auth_token()
    rest.handle_delete(f"/api/me/keys/{name}", success_msg="Successfully revoked")


@app.command()
def describe(file: str = typer.Option(..., help="File name")):
    """
    Show contents of specified access token file
    """
    print(read_file(file))


@app.command()
def activate(file: str = typer.Argument(..., help="File name")):
    """
    Activate a access token file to be used for authentication.

    --file

    This will be used as the current access token in all subsequent requests. This would
    be the same as login with a browser.
    """
    do_activate(file)


def do_activate(file: str):
    # Best effort to read file
    if os.path.exists(file):
        token_file = json.loads(read_file_with_path(file))
        write_token(token_file["token"])
    elif os.path.exists(str(Path.home()) + f"/.config/.remotive/{file}"):
        token_file = json.loads(read_file(file))
        write_token(token_file["token"])
    else:
        sys.stderr.write("File could not be found \n")


@app.command(name="list-files")
def list_files():
    """
    List personal access token files in remotivelabs config directory
    """
    personal_files = filter(lambda f: f.startswith("personal"), os.listdir(config_dir_name))
    for file in personal_files:
        print(file)


def read_file(file):
    f = open(str(Path.home()) + f"/.config/.remotive/{file}", "r")
    token = f.read()
    f.close()
    return token


def read_file_with_path(file):
    f = open(file, "r")
    token = f.read()
    f.close()
    return token


def write_token(token):
    f = open(token_file_name, "w")
    f.write(token)
    f.close()


def write_personal_token(file: str, token: str):
    path = str(Path.home()) + f"/.config/.remotive/{file}"
    f = open(path, "w")
    f.write(token)
    f.close()
    return path
